//
// knight-path - shortest knight path between two squares
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef KNIGHT_PATH_POSITION_H
#define KNIGHT_PATH_POSITION_H

#include <iostream>
#include <vector>

/**
  * The position class is used to represent a position on a chess board.
  * The board is assumed to be a standard 8x8 board.
  *
  * Input and output using algebraic notation is available.  See
  * http://en.wikipedia.org/wiki/Algebraic_notation_%28chess%29
  * for more information.
  *
  * Rows (called "ranks") are numbered 1..8, when viewed from white's
  * side of the board, with 1 closest (at the bottom) and 8 farthest (at
  * the top).
  *
  * Columns (called "files") are numbered 'a' to 'h', when viewed from
  * white's side of the board with 'a' on the left and 'h' on the right.
  */
class position
{
public:
    typedef std::vector<position> list_t;

    /**
      * @brief
      *     The destructor.
      *
      * It is not virtual, thou shalt not derive from this class.
      */
    ~position();

    /**
      * @brief
      *     The default constructor.
      *
      * The position is undefined, and will result in a run-time assert
      * failure if used without being initialised.
      */
    position();

    /**
      * @brief
      *     The default constructor.
      *
      * The position is undefined, and will result in a run-time assert
      * failure if used without being initialised.
      *
      * @param rank
      *     The Y coordinate, as a char between '1' and '8'.
      * @param file
      *     The X coordinate, as a char between 'a' and 'h' (or, to be
      *     generous, from 'A' to 'H').
      */
    position(char rank, char file);

    /**
      * The copy constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    position(const position &rhs);

    /**
      * The assignment operator.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    position &operator=(const position &rhs);

    /**
      * The knight_moves method is used to generate a list of legal
      * knight moves from this position.  It is assumed there are no
      * other pieces on the board.
      *
      * @returns
      *     A list of positions.
      *     There could be as many as 8, or a few as 2.
      */
    list_t knight_moves(void) const;

    /**
      * The read class method is used to read a position, in text form,
      * from an input stream.  Both upper and lower case are accepted for
      * the file/column portion.
      *
      * @param is
      *     The text input stream to be read.
      * @returns
      *     A valid position value
      * @note
      *     This class method does not return if there is an error
      *     (not present, malformed, eof-of-file, etc) instead a fatal
      *     error message will be printed, and the program will exit
      *     indicating failure.
      */
    static position read(std::istream &is);

    /**
      * The write method is used to write a text representation of a
      * chess board position.
      *
      * @param os
      *     The output stream to write the position onto.
      */
    void write(std::ostream &os) const;

    /**
      * The equality operator is used to test two position for equality.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true of the positions are the same, false if they are different.
      */
    bool operator==(const position &rhs) const;

    /**
      * The inequality operator is used to test two position for inequality.
      *
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     true of the positions are different, false if they are the same.
      */
    bool operator!=(const position &rhs) const;

    /**
      * The plot method is used to draw the given character into the
      * given board at this position.
      *
      * @param board
      *     The board to be drawn on.  The first domension is y|rank,
      *     the second dimension is x|file.
      * @param c
      *     The character to be drawn on the board.
      */
    void plot(char board[8][8], char c) const;

    /**
      * The distance method may be used to calculate the taxi-cab
      * distance (see http://en.wikipedia.org/wiki/Taxicab_geometry)
      * between this point and the given point.
      *
      * @param elsewhere
      *     The other position of interest.
      */
    int distance(const position &elsewhere) const;

private:
    /**
      * A constructor.
      * This is not part of the public interface.
      *
      * @param x
      *     The X coordinate of the position (0..7)
      * @param y
      *     The Y coordinate of the position (0..7)
      * @note
      *     The arguments are the reverse order to the char,char constructor.
      */
    position(int x, int y);

    /**
      * The p instance variable is used to remember the rank and file of
      * the position, in a single variable.
      *
      * @note
      *     The internal representation is undefined, and subject to
      *     change without notice.
      */
    unsigned p;

    /**
      * The valid method is used to ensure that the position is valid
      * before any processing takes place.
      */
    bool valid(void) const;

    /**
      * The get_rank method is used to obtain the row number.  Each row
      * (called a "rank" in chess jargon) are numbered 1..8, when viewed
      * from white's side of the board, with 1 closest (at the bottom)
      * and 8 farthest (at the top).
      *
      * @returns
      *     the rank, as a character from '1' to '8'
      */
    char get_rank(void) const;

    /**
      * The get_file method is used to obtain the column number.  Each
      * column (called a "file" in chess jargon) are numbered 'a' to
      * 'h', when viewed from white's side of the board with 'a' on the
      * left and 'h' on the right.
      *
      * @returns
      *     the file, as a character from 'a' to 'h'
      */
    char get_file(void) const;

public:
    // reluctantly made public, becaiuse solver_maze needs access

    /**
      * The get_x method is used to obtain the X coordinate of the position.
      *
      * @returns
      *     the X coordinate of the position (0..7)
      */
    int get_x(void) const;

    /**
      * The get_y method is used to obtain the Y coordinate of the position.
      *
      * @returns
      *     the Y coordinate of the position (0..7)
      */
    int get_y(void) const;
};

/**
  * The >> operator is used to read a position from an input stream.  It
  * will not return if there is an error reading the position (such as
  * not present, not valid, or end-of-file).
  *
  * @param is
  *     The input stream to be read.
  * @param p
  *     The resulting position value.
  */
inline std::istream &
operator>>(std::istream &is, position &p)
{
    p = position::read(is);
    return is;
}

/**
  * The << operator is used to write a position to an output stream.
  *
  * @param os
  *     The output stream to be written on.
  * @param p
  *     The position to be written.
  */
inline std::ostream &
operator<<(std::ostream &os, const position &p)
{
    p.write(os);
    return os;
}

#endif // KNIGHT_PATH_POSITION_H
