//
// knight-path - shortest knight path between two squares
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef KNIGHT_PATH_SOLVER_H
#define KNIGHT_PATH_SOLVER_H

#include <boost/shared_ptr.hpp>
#include <list>

#include <knight-path/position.h>

/**
  * The solver abstract base class is used to represent an unspecified
  * algorithm to solve the knight path problem in as few moves as possible.
  */
class solver
{
public:
    typedef boost::shared_ptr<solver> pointer;

    typedef std::list<position> list_t;

    /**
      * The destructor.
      */
    virtual ~solver();

public:
    /**
      * The factory class method is used to create new dynamically
      * allocated instances of the named solver class.
      *
      * @param name
      *     The name of the solver to be used.
      * @returns
      *     A pointer to a solver object.
      * @note
      *     If an unknown name is presented, a fatal error will be
      *     issued, and this class method will not return.
      */
    static pointer factory(const char *name);

    /**
      * The solve method is used to calculate the shortest path a
      * knight piece must take between two given positions on a chess
      * board.  There are often several equally short paths, the first
      * one found will be returned.
      *
      * @param from
      *     The position at the start of the path.
      * @param to
      *     The position at the end of the path.
      * @returns
      *     On success, a list of board positions will be returned;
      *       it will be a list of squares through which the Knight
      *       passes, the list will include the ending position, but
      *       will exclude the starting position (unless from == to).
      *     On error, (i.e. when no path is possible) an empty list
      *       will be returned.
      * @note
      *     Behaviour is undefined if either position is not valid.
      */
    virtual list_t solve(const position &from, const position &to) = 0;

    /**
      * The list_algorithms class method may be used to list the known
      * algorithm names.  This allows the user to choose an algorithm
      * name to pass to the "-a" option.
      */
    static void list_algorithms(void);

protected:
    /**
      * The default constructor.
      * For use by derived classes only.
      */
    solver();

private:
    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    solver(const solver &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    solver &operator=(const solver &rhs);
};

#endif // KNIGHT_PATH_SOLVER_H
