//
// knight-path - shortest knight path between two squares
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef KNIGHT_PATH_SOLVER_BRUTE_FORCE_H
#define KNIGHT_PATH_SOLVER_BRUTE_FORCE_H

#include <knight-path/solver.h>

/**
  * The solver_brute_force class is used to represent the processing
  * required to solve the knight path problem, using a brute force
  * recursive algorithm.
  *
  * The worst case is that there will be 7**6 paths explored (because 6
  * is the longest path) and the worst case usually occurs.  It would be
  * 8**6, except that some simple anti-backtracking is emplyed.
  *
  * Note that this is generally O(e**n) where n is the length of the
  * path, and would thus get *much* worse as the size of the board
  * increases.  This is because the recursive descent is a depth-first
  * search, and will likely discover longer paths before shorter ones.
  *
  * This still solves the problem in less than a second, on a relatively
  * modern laptop; faster than a human can do it.
  */
class solver_brute_force:
    public solver
{
public:
    /**
      * The destructor.
      */
    virtual ~solver_brute_force();

    // See base class for documentation.
    list_t solve(const position &from, const position &to);

    /**
      * The create class method is used to create new dynamically
      * allocated instances of this class.
      */
    static pointer create(void);

private:
    /**
      * The default constructor.
      * It is private on purpose, use a #create class method instead.
      */
    solver_brute_force();

    /**
      * The shortest_solution instance variable is used to remember
      * the shortest path found between the two points given to the
      * #solve method.  It is set by the #ply method when it finds a new
      * solution.
      */
    list_t shortest_solution;

    /**
      * The ply method is used to perform one ply of recursive analysis.
      * The shortest path found is store into the #shortest_solution
      * instance variable.
      *
      * @param from
      *     The position to start from
      * @param to
      *     The position to finish at.
      * @param progress
      *     The cumulative path so far, from earlier ply()s.
      */
    void ply(const position &from, const position &to, const list_t &progress);

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    solver_brute_force(const solver_brute_force &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    solver_brute_force &operator=(const solver_brute_force &rhs);
};

#endif // KNIGHT_PATH_SOLVER_BRUTE_FORCE_H
