//
// knight-path - shortest knight path between two squares
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef KNIGHT_PATH_SOLVER_MAZE_H
#define KNIGHT_PATH_SOLVER_MAZE_H

#include <knight-path/solver.h>

/**
  * The solver_maze class is used to represent solving the knigth
  * path problem using techniques more commonly found when building
  * singlely-connected mazes.
  *
  * Each square on the board is labeled with its distance from the
  * "from" square, and it predecessor.  If a new path is found to a
  * square, and it has a shorter distance, that new solution replaces
  * the existing solution.  Once all squares are visited, the solution
  * is constructed by tracing backwards from the "to" square.
  *
  * This algorithm runs in less than 0.7% of the time taken by the brute
  * force search.  This is because it only ever visits 64 squares, and
  * does no backtracking at all.
  *
  * In general, this algorithm is O(n) where n is the length of the
  * path.  This is because it does a breadth-first search for paths,
  * exploring the shortest paths before longer paths.  As a result, the
  * first path it finds will be the shortest (or equal-shortest, as
  * there are usually multiple solutions of any given length).  Given
  * that the length of the path is linear with the edge length of the
  * board, and *not* the square of the edge length, this is a huge
  * advantage over any algorithm that is polynomial in either path
  * length or edge length.
  */
class solver_maze:
    public solver
{
public:
    /**
      * The destructor.
      */
    virtual ~solver_maze();

    /**
      * The create class method is used to create new dynamically
      * allocated instance of the solver_maze
      * class.
      */
    static pointer create(void);

    // See base class for documentation.
    list_t solve(const position &from, const position &to);

private:
    /**
      * The default constructor.
      * It is private on purpose, use a #create class method instead.
      */
    solver_maze();

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    solver_maze(const solver_maze &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    solver_maze &operator=(const solver_maze &rhs);
};

#endif // KNIGHT_PATH_SOLVER_MAZE_H
