//
// knight-path - shortest knight path between two squares
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef KNIGHT_PATH_SOLVER_TAXI_H
#define KNIGHT_PATH_SOLVER_TAXI_H

#include <knight-path/solver.h>

/**
  * The solver_taxi class is used to represent the processing required
  * to solve the knight path problem, by sorting a superficially brute
  * force search by the distance from the goal.  Recursion depth
  * is limited by the shortest solution to date, thus eliminating
  * non-productive search directions.
  *
  * This algorithm performs in about 80% of the brute force time.
  */
class solver_taxi:
    public solver
{
public:
    /**
      * The destructor.
      */
    virtual ~solver_taxi();

    /**
      * The create class method is used to create new dynamically
      * allocated instance of the solver_taxi
      * class.
      */
    static pointer create(void);

    // See base class for documentation.
    list_t solve(const position &from, const position &to);

private:
    /**
      * The default constructor.
      * It is private on purpose, use a #create class method instead.
      */
    solver_taxi();

    /**
      * The shortest_solution instance variable is used to remember
      * the shortest path found between the two points given to the
      * #solve method.  It is set by the #ply method when it finds a new
      * solution.
      */
    list_t shortest_solution;

    /**
      * The ply method is used to perform one ply of recursive analysis.
      * The shortest path found is store into the #shortest_solution
      * instance variable.
      *
      * @param from
      *     The position to start from
      * @param to
      *     The position to finish at.
      * @param progress
      *     The cumulative path so far, from earlier ply()s.
      */
    void ply(const position &from, const position &to, const list_t &progress);

    /**
      * The copy constructor.
      * It is private on purpose, use a #create class method instead.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    solver_taxi(const solver_taxi &rhs);

    /**
      * The assignment operator.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    solver_taxi &operator=(const solver_taxi &rhs);
};

#endif // KNIGHT_PATH_SOLVER_TAXI_H
