//
// knight-path - shortest knight path between two squares
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef KNIGHT_PATH_TAXI_DISTANCE_H
#define KNIGHT_PATH_TAXI_DISTANCE_H

#include <knight-path/position.h>

/**
  * The taxi_distance class is used to represent the comparison to be
  * used when sorting a list of positions by their relative distances to
  * a goal position.
  *
  * This is used, for example, when sorting a list of potential knight moves:
  * @code
  * position::list_t candidates = from.knight_moves();
  * sort(candidates.begin(), candidates.end(), taxi_distance(to));
  * @endcode
  */
class taxi_distance
{
public:
    /**
      * The destructor.
      * It is not virtual, thou shalt not derive from it.
      */
    ~taxi_distance() { }

    /**
      * The constructor
      *
      * @param agoal
      *     The "to" position of interest.
      */
    taxi_distance(const position &agoal) : goal(agoal) { }

    /**
      * The function call operator is used by the sort algorithm to
      * compare two positions.
      *
      * @param lhs
      *     The left hand side of the "less than" comparison.
      * @param rhs
      *     The right hand side of the "less than" comparison.
      */
    bool
    operator()(const position &lhs, const position &rhs)
    {
        return (goal.distance(lhs) < goal.distance(rhs));
    }

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    taxi_distance(const taxi_distance &rhs) :
        goal(rhs.goal)
    {
    }

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    taxi_distance &operator=(const taxi_distance &rhs)
    {
        if (this != &rhs)
        {
            goal = rhs.goal;
        }
        return *this;
    }

private:
    /**
      * The goal instance variable is used to remember the "to"
      * position, when comparing two positions for their relative distance
      * to the goal.
      */
    position goal;

    /**
      * The default constructor.  Do not use.
      */
    taxi_distance();
};

#endif // KNIGHT_PATH_TAXI_DISTANCE_H
